from unittest import IsolatedAsyncioTestCase

from ......anoncreds.holder import AnonCredsHolder
from ......core.oob_processor import OobMessageProcessor
from ......indy.holder import IndyHolder
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......storage.error import StorageNotFoundError
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...formats.indy import handler as test_indy_handler
from ...messages.pres_format import V20PresFormat
from ...messages.pres_proposal import V20PresProposal
from ...messages.pres_request import V20PresRequest
from .. import pres_request_handler as test_module

S_ID = "NcYxiDXkpYi6ov5FcYDi1e:2:vidya:1.0"
CD_ID = f"NcYxiDXkpYi6ov5FcYDi1e:3:CL:{S_ID}:tag1"
INDY_PROOF_REQ = {
    "name": "proof-request",
    "version": "1.0",
    "nonce": "1234567890",
    "requested_attributes": {
        "0_favourite_uuid": {
            "name": "favourite",
            "restrictions": [
                {
                    "cred_def_id": CD_ID,
                }
            ],
        },
        "1_icon_uuid": {
            "name": "icon",
            "restrictions": [
                {
                    "cred_def_id": CD_ID,
                }
            ],
        },
    },
    "requested_predicates": {},
}
INDY_PROOF_REQ_PRED = {
    "name": "proof-request",
    "version": "1.0",
    "nonce": "1234567890",
    "requested_attributes": {},
    "requested_predicates": {
        "0_score_GE_uuid": {
            "name": "score",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [
                {
                    "cred_def_id": CD_ID,
                }
            ],
        }
    },
}
DIF_PROOF_REQ = {
    "options": {
        "challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "domain": "4jt78h47fh47",
    },
    "presentation_definition": {
        "id": "32f54163-7166-48f1-93d8-ff217bdb0653",
        "submission_requirements": [
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "count": 1,
                "from_nested": [
                    {
                        "name": "United States Citizenship Proofs",
                        "purpose": "We need you to prove you are a US citizen.",
                        "rule": "all",
                        "from": "A",
                    },
                    {
                        "name": "European Union Citizenship Proofs",
                        "purpose": "We need you to prove you are a citizen of a EU country.",
                        "rule": "all",
                        "from": "B",
                    },
                ],
            }
        ],
        "input_descriptors": [
            {
                "id": "citizenship_input_1",
                "name": "EU Driver's License",
                "group": ["A"],
                "schema": [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"}
                ],
                "constraints": {
                    "fields": [
                        {
                            "path": ["$.issuer.id", "$.issuer", "$.vc.issuer.id"],
                            "purpose": "The claim must be from one of the specified issuers",
                            "filter": {
                                "type": "string",
                                "enum": ["did:sov:4cLztgZYocjqTdAZM93t27"],
                            },
                        }
                    ]
                },
            },
            {
                "id": "citizenship_input_2",
                "name": "US Passport",
                "group": ["B"],
                "schema": [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"}
                ],
                "constraints": {
                    "fields": [
                        {
                            "path": ["$.issuanceDate", "$.vc.issuanceDate"],
                            "filter": {
                                "type": "string",
                                "format": "date",
                                "maximum": "2009-5-16",
                            },
                        }
                    ]
                },
            },
        ],
    },
}

DIF_PROP_REQ = {
    "input_descriptors": [
        {
            "id": "citizenship_input_1",
            "name": "EU Driver's License",
            "group": ["A"],
            "schema": [
                {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"}
            ],
            "constraints": {
                "fields": [
                    {
                        "path": ["$.issuer.id", "$.issuer", "$.vc.issuer.id"],
                        "purpose": "The claim must be from one of the specified issuers",
                        "filter": {
                            "type": "string",
                            "enum": ["did:sov:4cLztgZYocjqTdAZM93t27"],
                        },
                    }
                ]
            },
        },
        {
            "id": "citizenship_input_2",
            "name": "US Passport",
            "group": ["B"],
            "schema": [
                {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"}
            ],
            "constraints": {
                "fields": [
                    {
                        "path": ["$.issuanceDate", "$.vc.issuanceDate"],
                        "filter": {
                            "type": "string",
                            "format": "date",
                            "maximum": "2009-5-16",
                        },
                    }
                ]
            },
        },
    ]
}


class TestPresRequestHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()
        self.request_context = RequestContext.test_context(self.profile)
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.message = V20PresRequest()
        self.request_context.message.attachment = mock.MagicMock(
            return_value=mock.MagicMock()
        )

        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )

        self.mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)

    async def test_called(self):
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        px_rec_instance = test_module.V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock.MagicMock(auto_present=False)
            )

            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_not_found(self):
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        px_rec_instance = test_module.V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                side_effect=StorageNotFoundError
            )
            mock_px_rec_cls.return_value = px_rec_instance

            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock.MagicMock(auto_present=False)
            )

            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        assert not responder.messages

    async def test_called_auto_present_x_indy(self):
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        mock_px_rec = mock.MagicMock(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
            save_error_state=mock.CoroutineMock(),
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = mock_px_rec
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                side_effect=test_module.IndyHolderError()
            )

            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()

            with mock.patch.object(
                handler._logger, "exception", mock.MagicMock()
            ) as mock_log_exc:
                await handler.handle(self.request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_auto_present_x_anoncreds(self):
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        mock_px_rec = mock.MagicMock(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
            save_error_state=mock.CoroutineMock(),
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = mock_px_rec
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_pres = mock.AsyncMock(
                side_effect=test_module.AnonCredsHolderError()
            )

            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()

            with mock.patch.object(
                handler._logger, "exception", mock.MagicMock()
            ) as mock_log_exc:
                await handler.handle(self.request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_auto_present_indy(self):
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        mock_px_rec = test_module.V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = mock_px_rec
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(mock_px_rec, "pres message")
            )

            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()

            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            mock_px_rec
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_anoncreds(self):
        self.profile = await create_test_profile({"wallet.type": "askar-anoncreds"})
        self.request_context = RequestContext.test_context(self.profile)
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.connection_record.connection_id = "dummy"
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.message = V20PresRequest()
        self.request_context.message.attachment = mock.MagicMock(
            return_value=mock.MagicMock()
        )

        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )

        self.mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        mock_px_rec = test_module.V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = mock_px_rec
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(mock_px_rec, "pres message")
            )

            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()

            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            mock_px_rec
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_dif(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=V20PresFormat.Format.DIF.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=DIF_PROOF_REQ
        )

        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=V20PresFormat.Format.DIF.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_json(DIF_PROP_REQ, ident="dif")],
        )

        px_rec_instance = test_module.V20PresExRecord(
            pres_proposal=pres_proposal,
            auto_present=True,
        )
        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler_inst = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler_inst.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_no_preview_indy(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V20PresExRecord(auto_present=True)

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {"cred_info": {"referent": "dummy-0"}},
                    {"cred_info": {"referent": "dummy-1"}},
                ]
            )
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_no_preview_anoncreds(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V20PresExRecord(auto_present=True)

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {"cred_info": {"referent": "dummy-0"}},
                    {"cred_info": {"referent": "dummy-1"}},
                ]
            )
        )
        self.request_context.injector.bind_instance(AnonCredsHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_pred_no_match_indy(self):
        self.request_context.message = V20PresRequest()
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )
        self.request_context.message_receipt = MessageReceipt()
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        mock_px_rec = mock.MagicMock(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
            save_error_state=mock.CoroutineMock(),
        )

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[])
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = mock_px_rec
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                side_effect=test_indy_handler.V20PresFormatHandlerError
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()

            await handler.handle(self.request_context, responder)
            mock_px_rec.save_error_state.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            mock_px_rec
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )

    async def test_called_auto_present_pred_no_match_anoncreds(self):
        self.request_context.message = V20PresRequest()
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )
        self.request_context.message_receipt = MessageReceipt()
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )
        mock_px_rec = mock.MagicMock(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
            save_error_state=mock.CoroutineMock(),
        )

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[])
        )
        self.request_context.injector.bind_instance(AnonCredsHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = mock_px_rec
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock_px_rec
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                side_effect=test_indy_handler.V20PresFormatHandlerError
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()

            await handler.handle(self.request_context, responder)
            mock_px_rec.save_error_state.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            mock_px_rec
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )

    async def test_called_auto_present_pred_single_match_indy(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ_PRED
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V20PresExRecord(auto_present=True)

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy-0"}}])
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_pred_single_match_anoncreds(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ_PRED
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V20PresExRecord(auto_present=True)

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=[{"cred_info": {"referent": "dummy-0"}}])
        )
        self.request_context.injector.bind_instance(AnonCredsHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_pred_multi_match_indy(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ_PRED
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V20PresExRecord(auto_present=True)

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {"cred_info": {"referent": "dummy-0"}},
                    {"cred_info": {"referent": "dummy-1"}},
                ]
            )
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_pred_multi_match_anoncreds(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ_PRED
        )
        self.request_context.message_receipt = MessageReceipt()
        px_rec_instance = test_module.V20PresExRecord(auto_present=True)

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {"cred_info": {"referent": "dummy-0"}},
                    {"cred_info": {"referent": "dummy-1"}},
                ]
            )
        )
        self.request_context.injector.bind_instance(AnonCredsHolder, self.mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_multi_cred_match_reft_indy(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )
        self.request_context.message_receipt = MessageReceipt()
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {
                        "cred_info": {
                            "referent": "dummy-0",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "zero",
                                "favourite": "potato",
                                "icon": "cG90YXRv",
                            },
                        }
                    },
                    {
                        "cred_info": {
                            "referent": "dummy-1",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "one",
                                "favourite": "spud",
                                "icon": "c3B1ZA==",
                            },
                        }
                    },
                    {
                        "cred_info": {
                            "referent": "dummy-2",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "two",
                                "favourite": "patate",
                                "icon": "cGF0YXRl",
                            },
                        }
                    },
                ]
            )
        )
        self.request_context.injector.bind_instance(IndyHolder, self.mock_holder)

        px_rec_instance = test_module.V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
        )
        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_auto_present_multi_cred_match_reft_anoncreds(self):
        self.request_context.message = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ]
        )
        self.request_context.message.attachment = mock.MagicMock(
            return_value=INDY_PROOF_REQ
        )
        self.request_context.message_receipt = MessageReceipt()
        pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="indy",
                    format_=V20PresFormat.Format.INDY.aries,
                )
            ],
            proposals_attach=[AttachDecorator.data_base64(INDY_PROOF_REQ, ident="indy")],
        )

        self.mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(
                return_value=[
                    {
                        "cred_info": {
                            "referent": "dummy-0",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "zero",
                                "favourite": "potato",
                                "icon": "cG90YXRv",
                            },
                        }
                    },
                    {
                        "cred_info": {
                            "referent": "dummy-1",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "one",
                                "favourite": "spud",
                                "icon": "c3B1ZA==",
                            },
                        }
                    },
                    {
                        "cred_info": {
                            "referent": "dummy-2",
                            "cred_def_id": CD_ID,
                            "attrs": {
                                "ident": "two",
                                "favourite": "patate",
                                "icon": "cGF0YXRl",
                            },
                        }
                    },
                ]
            )
        )
        self.request_context.injector.bind_instance(AnonCredsHolder, self.mock_holder)

        px_rec_instance = test_module.V20PresExRecord(
            pres_proposal=pres_proposal.serialize(),
            auto_present=True,
        )
        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
        ):
            mock_pres_ex_rec_cls.return_value = px_rec_instance
            mock_pres_ex_rec_cls.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=px_rec_instance
            )
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_pres_mgr.return_value.create_pres = mock.CoroutineMock(
                return_value=(px_rec_instance, "pres message")
            )
            self.request_context.connection_ready = True
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)
            mock_pres_mgr.return_value.create_pres.assert_called_once()

        mock_pres_mgr.return_value.receive_pres_request.assert_called_once_with(
            px_rec_instance
        )
        self.mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            self.request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "pres message"
        assert target == {}

    async def test_called_not_ready(self):
        self.request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "V20PresManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_pres_request = mock.CoroutineMock()
            self.request_context.message = V20PresRequest()
            self.request_context.connection_ready = False
            handler = test_module.V20PresRequestHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(self.request_context, responder)
            assert (
                err.exception.message
                == "Connection used for presentation request not ready"
            )

        assert not responder.messages

        self.request_context = RequestContext.test_context(await create_test_profile())

        self.mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=None
        )
        self.request_context.injector.bind_instance(
            OobMessageProcessor, self.mock_oob_processor
        )

        self.request_context.message = V20PresRequest()
        handler = test_module.V20PresRequestHandler()
        responder = MockResponder()
        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(self.request_context, responder)
        assert (
            err.exception.message
            == "No connection or associated connectionless exchange found for presentation request"
        )

        assert not responder.messages



"""The please-ack decorator to request acknowledgement."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from ..models.base import BaseModel, BaseModelSchema
from ..valid import UUID4_EXAMPLE


class PleaseAckDecorator(BaseModel):
    """Class representing the please-ack decorator."""

    class Meta:
        """PleaseAckDecorator metadata."""

        schema_class = "PleaseAckDecoratorSchema"

    def __init__(
        self,
        message_id: Optional[str] = None,
        on: Sequence[str] = None,
    ):
        """Initialize a PleaseAckDecorator instance.

        Args:
            message_id: identifier of message to acknowledge, if not current message
            on: list of tokens describing circumstances for acknowledgement.

        """
        super().__init__()
        self.message_id = message_id
        self.on = list(on) if on else None


class PleaseAckDecoratorSchema(BaseModelSchema):
    """PleaseAck decorator schema used in serialization/deserialization."""

    class Meta:
        """PleaseAckDecoratorSchema metadata."""

        model_class = PleaseAckDecorator
        unknown = EXCLUDE

    message_id = fields.Str(
        required=False,
        allow_none=False,
        metadata={"description": "Message identifier", "example": UUID4_EXAMPLE},
    )
    on = fields.List(
        fields.Str(metadata={"example": "OUTCOME"}),
        required=False,
        metadata={
            "description": "List of tokens describing circumstances for acknowledgement"
        },
    )



import json

import indy_vdr
import pytest
import pytest_asyncio

from ...anoncreds.default.legacy_indy.registry import LegacyIndyRegistry
from ...cache.base import BaseCache
from ...cache.in_memory import InMemoryCache
from ...indy.issuer import IndyIssuer
from ...tests import mock
from ...utils.testing import create_test_profile
from ...wallet.base import BaseWallet
from ...wallet.did_info import DIDInfo
from ...wallet.did_method import SOV, DIDMethod, DIDMethods, HolderDefinedDid
from ...wallet.did_posture import DIDPosture
from ...wallet.key_type import ED25519, KeyTypes
from ..endpoint_type import EndpointType
from ..indy_vdr import (
    BadLedgerRequestError,
    ClosedPoolError,
    IndyVdrLedger,
    IndyVdrLedgerPool,
    LedgerError,
    LedgerTransactionError,
    Role,
    VdrError,
)

WEB = DIDMethod(
    name="web",
    key_types=[ED25519],
    rotation=True,
    holder_defined_did=HolderDefinedDid.REQUIRED,
)


@pytest_asyncio.fixture
async def ledger():
    did_methods = DIDMethods()
    did_methods.register(WEB)
    profile = await create_test_profile()
    profile.context.injector.bind_instance(DIDMethods, did_methods)
    profile.context.injector.bind_instance(BaseCache, InMemoryCache())
    profile.context.injector.bind_instance(KeyTypes, KeyTypes())

    ledger = IndyVdrLedger(IndyVdrLedgerPool("test-ledger"), profile)

    async def open():
        ledger.pool.handle = mock.MagicMock(indy_vdr.Pool)

    async def close():
        ledger.pool.handle = None

    with (
        mock.patch.object(ledger.pool, "open", open),
        mock.patch.object(ledger.pool, "close", close),
        mock.patch.object(
            ledger, "is_ledger_read_only", mock.CoroutineMock(return_value=False)
        ),
    ):
        yield ledger


@pytest.mark.indy_vdr
class TestIndyVdrLedger:
    @pytest.mark.asyncio
    async def test_aenter_aexit(self, ledger: IndyVdrLedger):
        assert ledger.pool_handle is None

        async with ledger as led:
            assert ledger.pool_handle

        assert ledger.pool_handle is None

    @pytest.mark.asyncio
    async def test_submit_pool_closed(self, ledger: IndyVdrLedger):
        with pytest.raises(ClosedPoolError):
            await ledger._submit("{}")

    @pytest.mark.asyncio
    async def test_submit_signed(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            test_did = await wallet.create_public_did(SOV, ED25519)
            test_msg = indy_vdr.ledger.build_get_txn_request(test_did.did, 1, 1)

        async with ledger:
            result = await ledger._submit(
                test_msg, sign=True, taa_accept=False, write_ledger=False
            )
            assert result.get("signature")
            assert result.get("taaAcceptance") is None

            result = await ledger._submit(
                test_msg, sign=True, taa_accept=True, write_ledger=False
            )
            assert result.get("signature")
            assert result.get("taaAcceptance") is None

            ledger.pool_handle.submit_request.assert_not_awaited()
            await ledger._submit(test_msg)
            ledger.pool_handle.submit_request.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_fetch_txn_author_agreement(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            with mock.patch.object(
                ledger.pool_handle,
                "submit_request",
                mock.CoroutineMock(
                    side_effect=[
                        {"data": {"aml": ".."}},
                        {"data": {"text": "text", "version": "1.0"}},
                    ]
                ),
            ):
                response = await ledger.fetch_txn_author_agreement()
                assert response == {
                    "aml_record": {"aml": ".."},
                    "taa_record": {
                        "text": "text",
                        "version": "1.0",
                        "digest": ledger.taa_digest("1.0", "text"),
                    },
                    "taa_required": True,
                }

    @pytest.mark.asyncio
    async def test_submit_signed_taa_accept(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            test_did = await wallet.create_public_did(SOV, ED25519)

        async with ledger:
            test_msg = indy_vdr.ledger.build_get_txn_request(test_did.did, 1, 1)

            await ledger.accept_txn_author_agreement(
                {
                    "text": "txt",
                    "version": "ver",
                    "digest": ledger.taa_digest("ver", "txt"),
                },
                mechanism="manual",
                accept_time=1000,
            )

            result = await ledger._submit(
                test_msg, sign=True, taa_accept=True, write_ledger=False
            )
            assert result.get("signature")
            assert result.get("taaAcceptance")

            # no accept_time
            await ledger.accept_txn_author_agreement(
                {
                    "text": "txt",
                    "version": "ver",
                    "digest": ledger.taa_digest("ver", "txt"),
                },
                mechanism="manual",
            )

    @pytest.mark.asyncio
    async def test_submit_unsigned(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            test_msg = indy_vdr.ledger.build_get_txn_request(None, 1, 1)

            result = await ledger._submit(
                test_msg, sign=False, taa_accept=False, write_ledger=False
            )
            assert result.get("signature") is None
            assert result.get("taaAcceptance") is None

            # no public DID
            with pytest.raises(BadLedgerRequestError):
                await ledger._submit(test_msg, sign=True, write_ledger=False)

    @pytest.mark.asyncio
    async def test_submit_unsigned_ledger_transaction_error(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            test_msg = indy_vdr.ledger.build_get_txn_request(None, 1, 1)

            ledger.pool_handle.submit_request.side_effect = VdrError(99, "message")
            with pytest.raises(LedgerTransactionError):
                await ledger._submit(test_msg, sign=False, taa_accept=False)

    @pytest.mark.asyncio
    async def test_txn_endorse(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            test_msg = indy_vdr.ledger.build_get_txn_request(None, 1, 1)

            async with ledger:
                # invalid request
                with pytest.raises(BadLedgerRequestError):
                    await ledger.txn_endorse(request_json="{}")

                # no public DID
                with pytest.raises(BadLedgerRequestError):
                    await ledger.txn_endorse(request_json=test_msg.body)

                test_did = await wallet.create_public_did(SOV, ED25519)
                test_msg.set_endorser(test_did.did)

                endorsed_json = await ledger.txn_endorse(request_json=test_msg.body)
                body = json.loads(endorsed_json)
                assert test_did.did in body["signatures"]

    @pytest.mark.asyncio
    async def test_send_schema(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            test_did = await wallet.create_public_did(SOV, ED25519)
            issuer = mock.MagicMock(IndyIssuer)
            issuer.create_schema.return_value = (
                "schema_issuer_did:schema_name:9.1",
                (
                    r'{"ver": "1.0", "id": "schema_issuer_did:schema_name:9.1", "name":'
                    r' "schema_name", "version": "9.1", "attrNames": ["a", "b"]}'
                ),
            )

        async with ledger:
            ledger.pool_handle.submit_request.return_value = {"txnMetadata": {"seqNo": 1}}

            with mock.patch.object(
                ledger,
                "check_existing_schema",
                mock.CoroutineMock(return_value=None),
            ):
                schema_id, _ = await ledger.create_and_send_schema(
                    issuer, "schema_name", "9.1", ["a", "b"]
                )

            issuer.create_schema.assert_awaited_once_with(
                test_did.did, "schema_name", "9.1", ["a", "b"]
            )

            ledger.pool_handle.submit_request.assert_awaited_once()

            assert schema_id == issuer.create_schema.return_value[0]

            # test endorsed
            schema_id, signed_txn = await ledger.create_and_send_schema(
                issuer=issuer,
                schema_name="schema_name",
                schema_version="9.1",
                attribute_names=["a", "b"],
                write_ledger=False,
                endorser_did=test_did.did,
            )
            assert schema_id == issuer.create_schema.return_value[0]
            txn = json.loads(signed_txn["signed_txn"])
            assert txn.get("endorser") == test_did.did
            assert txn.get("signature")

    @pytest.mark.asyncio
    async def test_send_schema_no_public_did(
        self,
        ledger: IndyVdrLedger,
    ):
        issuer = mock.MagicMock(IndyIssuer)
        async with ledger:
            with pytest.raises(BadLedgerRequestError):
                await ledger.create_and_send_schema(
                    issuer, "schema_name", "9.1", ["a", "b"]
                )

    @pytest.mark.asyncio
    async def test_send_schema_already_exists(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        issuer = mock.MagicMock(IndyIssuer)
        issuer.create_schema.return_value = (
            "schema_issuer_did:schema_name:9.1",
            (
                r'{"ver": "1.0", "id": "schema_issuer_did:schema_name:9.1", "name":'
                r' "schema_name", "version": "9.1", "attrNames": ["a", "b"]}'
            ),
        )

        async with ledger:
            with mock.patch.object(
                ledger,
                "check_existing_schema",
                mock.CoroutineMock(
                    return_value=(
                        issuer.create_schema.return_value[0],
                        {"schema": "result"},
                    )
                ),
            ) as mock_check:
                schema_id, schema_def = await ledger.create_and_send_schema(
                    issuer, "schema_name", "9.1", ["a", "b"]
                )
                assert schema_id == mock_check.return_value[0]
                assert schema_def == mock_check.return_value[1]

    @pytest.mark.asyncio
    async def test_send_schema_ledger_read_only(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        issuer = mock.MagicMock(IndyIssuer)
        issuer.create_schema.return_value = (
            "schema_issuer_did:schema_name:9.1",
            (
                r'{"ver": "1.0", "id": "schema_issuer_did:schema_name:9.1", "name":'
                r' "schema_name", "version": "9.1", "attrNames": ["a", "b"]}'
            ),
        )

        async with ledger:
            ledger.pool.read_only = True
            with (
                mock.patch.object(
                    ledger,
                    "check_existing_schema",
                    mock.CoroutineMock(return_value=False),
                ),
                mock.patch.object(
                    ledger,
                    "is_ledger_read_only",
                    mock.CoroutineMock(return_value=True),
                ),
            ):
                with pytest.raises(LedgerError):
                    await ledger.create_and_send_schema(
                        issuer, "schema_name", "9.1", ["a", "b"]
                    )

    @pytest.mark.asyncio
    async def test_send_schema_ledger_transaction_error(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        issuer = mock.MagicMock(IndyIssuer)
        issuer.create_schema.return_value = (
            "schema_issuer_did:schema_name:9.1",
            (
                r'{"ver": "1.0", "id": "schema_issuer_did:schema_name:9.1", "name":'
                r' "schema_name", "version": "9.1", "attrNames": ["a", "b"]}'
            ),
        )

        async with ledger:
            ledger.pool_handle.submit_request.side_effect = VdrError(99, "message")

            with mock.patch.object(
                ledger,
                "check_existing_schema",
                mock.CoroutineMock(return_value=False),
            ):
                with pytest.raises(LedgerTransactionError):
                    await ledger.create_and_send_schema(
                        issuer, "schema_name", "9.1", ["a", "b"]
                    )

    @pytest.mark.asyncio
    async def test_send_schema_no_indy_did(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            wallet.create_public_did = mock.CoroutineMock(
                return_value={
                    "result": {
                        "did": "did:web:doma.in",
                        "verkey": "verkey",
                        "posture": DIDPosture.PUBLIC.moniker,
                        "key_type": ED25519.key_type,
                        "method": WEB.method_name,
                    }
                }
            )
        issuer = mock.MagicMock(IndyIssuer)
        async with ledger:
            with pytest.raises(BadLedgerRequestError):
                await ledger.create_and_send_schema(
                    issuer, "schema_name", "9.1", ["a", "b"]
                )

    @pytest.mark.asyncio
    async def test_get_schema(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "seqNo": 99,
                "dest": "55GkHamhTU1ZbTbV2ab9DE",
                "data": {
                    "name": "schema_name",
                    "version": "9.1",
                    "attr_names": ["a", "b"],
                },
            }

            result = await ledger.get_schema("55GkHamhTU1ZbTbV2ab9DE:2:schema_name:9.1")
            assert result == {
                "ver": "1.0",
                "id": "55GkHamhTU1ZbTbV2ab9DE:2:schema_name:9.1",
                "name": "schema_name",
                "version": "9.1",
                "attrNames": ["a", "b"],
                "seqNo": 99,
            }

    @pytest.mark.asyncio
    async def test_get_schema_not_found(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {}
            result = await ledger.get_schema("55GkHamhTU1ZbTbV2ab9DE:2:schema_name:9.1")
            assert result is None

    @pytest.mark.asyncio
    async def test_send_credential_definition(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        schema_id = "55GkHamhTU1ZbTbV2ab9DE:2:schema_name:9.1"
        cred_def_id = "55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag"
        cred_def = {
            "ver": "1.0",
            "id": cred_def_id,
            "schemaId": schema_id,
            "type": "CL",
            "tag": "tag",
            "value": {
                "primary": {
                    "n": "...",
                    "s": "...",
                    "r": "...",
                    "revocation": None,
                }
            },
        }
        issuer = mock.MagicMock(IndyIssuer)
        issuer.make_credential_definition_id.return_value = cred_def_id
        issuer.credential_definition_in_wallet.return_value = False
        issuer.create_and_store_credential_definition.return_value = (
            cred_def_id,
            json.dumps(cred_def),
        )

        async with ledger:
            ledger.pool_handle.submit_request.side_effect = (
                {
                    "seqNo": 99,
                    "dest": "55GkHamhTU1ZbTbV2ab9DE",
                    "data": {
                        "name": "schema_name",
                        "version": "9.1",
                        "attr_names": ["a", "b"],
                    },
                },
                {"data": None},  # cred def lookup result
                {},  # submission result
            )

            result = await ledger.create_and_send_credential_definition(
                issuer=issuer,
                schema_id=schema_id,
                signature_type="CL",
                tag="tag",
                support_revocation=False,
            )
            assert result == (cred_def_id, cred_def, True)

    @pytest.mark.asyncio
    async def test_send_credential_definition_no_public_did(
        self,
        ledger: IndyVdrLedger,
    ):
        issuer = mock.MagicMock(IndyIssuer)
        async with ledger:
            with pytest.raises(BadLedgerRequestError):
                await ledger.create_and_send_credential_definition(
                    issuer, "schema_id", None, "tag"
                )

    @pytest.mark.asyncio
    async def test_send_credential_definition_no_such_schema(self, ledger: IndyVdrLedger):
        issuer = mock.MagicMock(IndyIssuer)
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {}
            with pytest.raises(BadLedgerRequestError):
                await ledger.create_and_send_credential_definition(
                    issuer, "schema_id", None, "tag"
                )

    @pytest.mark.asyncio
    async def test_send_credential_definition_read_only(self, ledger: IndyVdrLedger):
        issuer = mock.MagicMock(IndyIssuer)
        async with ledger:
            ledger.pool.read_only = True
            with pytest.raises(LedgerError):
                await ledger.create_and_send_credential_definition(
                    issuer, "schema_id", None, "tag"
                )

    @pytest.mark.asyncio
    async def test_get_credential_definition(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "seqNo": 99,
                "ref": "schema-id",
                "signature_type": "CL",
                "tag": "tag",
                "origin": "origin-did",
                "data": {"cred": "def"},
            }

            result = await ledger.get_credential_definition(
                "55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag"
            )
            assert result == {
                "ver": "1.0",
                "id": "origin-did:3:CL:schema-id:tag",
                "schemaId": "schema-id",
                "type": "CL",
                "tag": "tag",
                "value": {"cred": "def"},
            }

    @pytest.mark.asyncio
    async def test_get_credential_definition_not_found(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "seqNo": 99,
                "ref": "schema-id",
                "signature_type": "CL",
                "tag": "tag",
                "origin": "origin-did",
                "data": None,
            }

            result = await ledger.get_credential_definition(
                "55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag"
            )
            assert result is None

    @pytest.mark.asyncio
    async def test_get_key_for_did(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"verkey": "VK"}',
            }
            result = await ledger.get_key_for_did("55GkHamhTU1ZbTbV2ab9DE")
            assert result == "VK"

    @pytest.mark.asyncio
    async def test_get_key_for_did_non_sov_public_did(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            async with ledger.profile.session() as session:
                wallet = session.inject(BaseWallet)
                wallet.get_public_did = mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did:web:doma.in",
                        "verkey",
                        DIDPosture.PUBLIC.metadata,
                        WEB,
                        ED25519,
                    )
                )
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"verkey": "VK"}',
            }
            result = await ledger.get_key_for_did("55GkHamhTU1ZbTbV2ab9DE")
            assert result == "VK"

    @pytest.mark.asyncio
    async def test_get_all_endpoints_for_did(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"endpoint": {"default": "endp"}}',
            }
            result = await ledger.get_all_endpoints_for_did("55GkHamhTU1ZbTbV2ab9DE")
            assert result == {"default": "endp"}

    @pytest.mark.asyncio
    async def test_get_all_endpoints_for_did_none(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": None,
            }
            result = await ledger.get_all_endpoints_for_did("55GkHamhTU1ZbTbV2ab9DE")
            assert result is None

    @pytest.mark.asyncio
    async def test_get_endpoint_for_did(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"endpoint": {"endpoint": "endp"}}',
            }
            result = await ledger.get_endpoint_for_did(
                "55GkHamhTU1ZbTbV2ab9DE", EndpointType.ENDPOINT
            )
            assert result == "endp"

    @pytest.mark.asyncio
    async def test_get_endpoint_for_did_address_none(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"endpoint": null}',
            }
            result = await ledger.get_endpoint_for_did(
                "55GkHamhTU1ZbTbV2ab9DE", EndpointType.ENDPOINT
            )
            assert result is None

    @pytest.mark.asyncio
    async def test_get_endpoint_for_did_empty(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": None,
            }
            result = await ledger.get_endpoint_for_did(
                "55GkHamhTU1ZbTbV2ab9DE", EndpointType.ENDPOINT
            )
            assert result is None

    @pytest.mark.asyncio
    async def test_update_endpoint_for_did(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            ledger.pool_handle.submit_request.side_effect = (
                {"data": None},
                {"data": None},
            )
            await ledger.update_endpoint_for_did(
                "55GkHamhTU1ZbTbV2ab9DE", "https://url", EndpointType.ENDPOINT
            )

    @pytest.mark.parametrize(
        "all_exist_endpoints, routing_keys, result",
        [
            (
                {"profile": "https://endpoint/profile"},
                ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
                {
                    "endpoint": {
                        "profile": "https://endpoint/profile",
                        "endpoint": "https://url",
                        "routingKeys": ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
                    }
                },
            ),
            (
                {"profile": "https://endpoint/profile"},
                None,
                {
                    "endpoint": {
                        "profile": "https://endpoint/profile",
                        "endpoint": "https://url",
                        "routingKeys": [],
                    }
                },
            ),
            (
                None,
                ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
                {
                    "endpoint": {
                        "endpoint": "https://url",
                        "routingKeys": ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
                    }
                },
            ),
            (None, None, {"endpoint": {"endpoint": "https://url", "routingKeys": []}}),
            (
                {
                    "profile": "https://endpoint/profile",
                    "spec_divergent_endpoint": "https://endpoint",
                },
                ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
                {
                    "endpoint": {
                        "profile": "https://endpoint/profile",
                        "spec_divergent_endpoint": "https://endpoint",
                        "endpoint": "https://url",
                        "routingKeys": ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
                    }
                },
            ),
        ],
    )
    @pytest.mark.asyncio
    async def test_construct_attr_json(
        self, ledger: IndyVdrLedger, all_exist_endpoints, routing_keys, result
    ):
        async with ledger:
            attr_json = await ledger._construct_attr_json(
                "https://url", EndpointType.ENDPOINT, all_exist_endpoints, routing_keys
            )
        assert attr_json == json.dumps(result)

    @pytest.mark.asyncio
    async def test_update_endpoint_for_did_calls_attr_json(self, ledger: IndyVdrLedger):
        routing_keys = ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"]
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            test_did = await wallet.create_public_did(SOV, ED25519)

        async with ledger:
            with (
                mock.patch.object(
                    ledger,
                    "_construct_attr_json",
                    mock.CoroutineMock(
                        return_value=json.dumps(
                            {
                                "endpoint": {
                                    "endpoint": {
                                        "endpoint": "https://url",
                                        "routingKeys": [],
                                    }
                                }
                            }
                        )
                    ),
                ) as mock_construct_attr_json,
                mock.patch.object(
                    ledger,
                    "get_all_endpoints_for_did",
                    mock.CoroutineMock(return_value={}),
                ),
            ):
                await ledger.update_endpoint_for_did(
                    test_did.did,
                    "https://url",
                    EndpointType.ENDPOINT,
                    routing_keys=routing_keys,
                )
                mock_construct_attr_json.assert_called_once_with(
                    "https://url",
                    EndpointType.ENDPOINT,
                    {},
                    routing_keys,
                )

    @pytest.mark.asyncio
    async def test_update_endpoint_for_did_no_public(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {"data": None}
            with pytest.raises(BadLedgerRequestError):
                await ledger.update_endpoint_for_did(
                    "55GkHamhTU1ZbTbV2ab9DE", "https://url", EndpointType.ENDPOINT
                )

    @pytest.mark.asyncio
    async def test_update_endpoint_for_did_read_only(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool.read_only = True
            ledger.pool_handle.submit_request.return_value = {"data": None}
            with pytest.raises(LedgerError):
                await ledger.update_endpoint_for_did(
                    "55GkHamhTU1ZbTbV2ab9DE", "https://url", EndpointType.ENDPOINT
                )

    @pytest.mark.asyncio
    async def test_register_nym_local(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
            post_did = await wallet.create_local_did(SOV, ED25519)
            async with ledger:
                await ledger.register_nym(post_did.did, post_did.verkey)
            did = await wallet.get_local_did(post_did.did)
            assert did.metadata["posted"] is True

    @pytest.mark.asyncio
    async def test_register_nym_non_local(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            await ledger.register_nym("55GkHamhTU1ZbTbV2ab9DE", "verkey")

    @pytest.mark.asyncio
    async def test_register_nym_no_public(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            with pytest.raises(BadLedgerRequestError):
                await ledger.register_nym("55GkHamhTU1ZbTbV2ab9DE", "verkey")

    @pytest.mark.asyncio
    async def test_register_nym_read_only(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool.read_only = True
            with pytest.raises(LedgerError):
                await ledger.register_nym(
                    "55GkHamhTU1ZbTbV2ab9DE",
                    "verkey",
                    "",
                )

    @pytest.mark.asyncio
    async def test_get_nym_role(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"role":"TRUSTEE"}'
            }
            result = await ledger.get_nym_role(
                "55GkHamhTU1ZbTbV2ab9DE",
            )
            assert result == Role.TRUSTEE

    @pytest.mark.asyncio
    async def test_get_nym_role_unknown(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {"data": r"{}"}
            with pytest.raises(BadLedgerRequestError):
                await ledger.get_nym_role(
                    "55GkHamhTU1ZbTbV2ab9DE",
                )

    @pytest.mark.asyncio
    async def test_get_nym_role_invalid(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            ledger.pool_handle.submit_request.return_value = {
                "data": r'{"role":"INVALID"}'
            }
            result = await ledger.get_nym_role(
                "55GkHamhTU1ZbTbV2ab9DE",
            )
            assert result is None

    @pytest.mark.asyncio
    async def test_get_revoc_reg_def(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            ledger.pool_handle.submit_request.return_value = {
                "data": {"id": reg_id},
                "txnTime": 1234567890,
            }
            result = await ledger.get_revoc_reg_def(
                reg_id,
            )
            assert result["id"] == reg_id
            assert result["txnTime"] == 1234567890

    @pytest.mark.asyncio
    async def test_get_revoc_reg_entry(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            ledger.pool_handle.submit_request.return_value = {
                "data": {
                    "id": reg_id,
                    "txnTime": 1234567890,
                    "value": "...",
                    "revocRegDefId": reg_id,
                },
            }
            result = await ledger.get_revoc_reg_entry(reg_id, 1234567890)
            assert result == ({"ver": "1.0", "value": "..."}, 1234567890)

    @pytest.mark.asyncio
    async def test_get_revoc_reg_delta(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            ledger.pool_handle.submit_request.return_value = {
                "data": {
                    "value": {
                        "accum_to": {
                            "value": {"accum": "ACCUM"},
                            "txnTime": 1234567890,
                        },
                        "issued": [1, 2],
                        "revoked": [3, 4],
                    },
                    "revocRegDefId": reg_id,
                },
            }
            result = await ledger.get_revoc_reg_delta(reg_id)
            assert result == (
                {
                    "ver": "1.0",
                    "value": {"accum": "ACCUM", "issued": [1, 2], "revoked": [3, 4]},
                },
                1234567890,
            )

    @pytest.mark.asyncio
    async def test_get_revoc_reg_delta_without_accum_to(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            ledger.pool_handle.submit_request.side_effect = [
                # First call to get_revoc_reg_delta
                {
                    "data": {
                        "value": {},
                        "revocRegDefId": reg_id,
                    },
                },
                # Get registry with test_get_revoc_reg_entry
                {
                    "data": {
                        "id": reg_id,
                        "txnTime": 1234567890,
                        "value": "...",
                        "revocRegDefId": reg_id,
                    },
                },
                # Second call to get_revoc_reg_delta
                {
                    "data": {
                        "value": {
                            "accum_to": {
                                "value": {"accum": "ACCUM"},
                                "txnTime": 1234567890,
                            },
                            "issued": [1, 2],
                            "revoked": [3, 4],
                        },
                        "revocRegDefId": reg_id,
                    },
                },
            ]
            result = await ledger.get_revoc_reg_delta(reg_id)
            assert result == (
                {
                    "ver": "1.0",
                    "value": {"accum": "ACCUM", "issued": [1, 2], "revoked": [3, 4]},
                },
                1234567890,
            )

    @pytest.mark.asyncio
    async def test_send_revoc_reg_def(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            reg_def = {
                "ver": "1.0",
                "id": reg_id,
                "revocDefType": "CL_ACCUM",
                "tag": "tag1",
                "credDefId": "55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag",
                "value": {
                    "issuanceType": "ISSUANCE_ON_DEMAND",
                    "maxCredNum": 5,
                    "publicKeys": {"accumKey": {"z": "1 ..."}},
                    "tailsHash": "",
                    "tailsLocation": "",
                },
            }
            ledger.pool_handle.submit_request.return_value = {"status": "ok"}
            result = await ledger.send_revoc_reg_def(reg_def, issuer_did=None)
            assert result == {"result": {"status": "ok"}}

    @pytest.mark.asyncio
    async def test_send_revoc_reg_def_anoncreds_write_to_ledger(
        self,
        ledger: IndyVdrLedger,
    ):
        ledger.profile.settings.set_value("wallet.type", "askar-anoncreds")
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            reg_def = {
                "ver": "1.0",
                "id": reg_id,
                "revocDefType": "CL_ACCUM",
                "tag": "tag1",
                "credDefId": "55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag",
                "value": {
                    "issuanceType": "ISSUANCE_ON_DEMAND",
                    "maxCredNum": 5,
                    "publicKeys": {"accumKey": {"z": "1 ..."}},
                    "tailsHash": "",
                    "tailsLocation": "",
                },
            }

            with mock.patch.object(
                LegacyIndyRegistry,
                "txn_submit",
                return_value=json.dumps({"result": {"txnMetadata": {"seqNo": 1234}}}),
            ):
                ledger.pool_handle.submit_request.return_value = {"status": "ok"}
                result = await ledger.send_revoc_reg_def(reg_def, issuer_did=None)
                assert result == 1234

    @pytest.mark.asyncio
    async def test_send_revoc_reg_def_anoncreds_do_not_write_to_ledger(
        self,
        ledger: IndyVdrLedger,
    ):
        ledger.profile.settings.set_value("wallet.type", "askar-anoncreds")
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            reg_def = {
                "ver": "1.0",
                "id": reg_id,
                "revocDefType": "CL_ACCUM",
                "tag": "tag1",
                "credDefId": "55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag",
                "value": {
                    "issuanceType": "ISSUANCE_ON_DEMAND",
                    "maxCredNum": 5,
                    "publicKeys": {"accumKey": {"z": "1 ..."}},
                    "tailsHash": "",
                    "tailsLocation": "",
                },
            }

            with mock.patch.object(
                LegacyIndyRegistry,
                "txn_submit",
                return_value=json.dumps({"result": {"txnMetadata": {"seqNo": 1234}}}),
            ):
                ledger.pool_handle.submit_request.return_value = {"status": "ok"}
                result = await ledger.send_revoc_reg_def(
                    reg_def, issuer_did=None, write_ledger=False
                )
                assert isinstance(result, tuple)

    @pytest.mark.asyncio
    async def test_send_revoc_reg_entry(
        self,
        ledger: IndyVdrLedger,
    ):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            reg_entry = {
                "ver": "1.0",
                "value": {},
            }
            ledger.pool_handle.submit_request.return_value = {"status": "ok"}
            result = await ledger.send_revoc_reg_entry(reg_id, "CL_ACCUM", reg_entry)
            assert result == {"result": {"status": "ok"}}

    @pytest.mark.asyncio
    async def test_send_revoc_reg_entry_anoncreds_write_to_ledger(
        self,
        ledger: IndyVdrLedger,
    ):
        ledger.profile.settings.set_value("wallet.type", "askar-anoncreds")
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            reg_entry = {
                "ver": "1.0",
                "value": {},
            }
            with mock.patch.object(
                LegacyIndyRegistry,
                "txn_submit",
                return_value=json.dumps({"result": {"txnMetadata": {"seqNo": 1234}}}),
            ):
                ledger.pool_handle.submit_request.return_value = {"status": "ok"}
                result = await ledger.send_revoc_reg_entry(
                    reg_id, "CL_ACCUM", reg_entry, write_ledger=False
                )
                assert isinstance(result, tuple)

    @pytest.mark.asyncio
    async def test_send_revoc_reg_entry_anoncreds_do_not_write_to_ledger(
        self,
        ledger: IndyVdrLedger,
    ):
        ledger.profile.settings.set_value("wallet.type", "askar-anoncreds")
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)
        async with ledger:
            reg_id = (
                "55GkHamhTU1ZbTbV2ab9DE:4:55GkHamhTU1ZbTbV2ab9DE:3:CL:99:tag:CL_ACCUM:0"
            )
            reg_entry = {
                "ver": "1.0",
                "value": {},
            }
            with mock.patch.object(
                LegacyIndyRegistry,
                "txn_submit",
                return_value=json.dumps({"result": {"txnMetadata": {"seqNo": 1234}}}),
            ):
                ledger.pool_handle.submit_request.return_value = {"status": "ok"}
                result = await ledger.send_revoc_reg_entry(reg_id, "CL_ACCUM", reg_entry)
                assert result == 1234

    @pytest.mark.asyncio
    async def test_credential_definition_id2schema_id(self, ledger: IndyVdrLedger):
        S_ID = "55GkHamhTU1ZbTbV2ab9DE:2:favourite_drink:1.0"
        SEQ_NO = "9999"

        async with ledger:
            with mock.patch.object(
                ledger,
                "get_schema",
                mock.CoroutineMock(return_value={"id": S_ID}),
            ) as mock_get_schema:
                s_id_short = await ledger.credential_definition_id2schema_id(
                    f"55GkHamhTU1ZbTbV2ab9DE:3:CL:{SEQ_NO}:tag"
                )

                mock_get_schema.assert_called_once_with(SEQ_NO)

                assert s_id_short == S_ID
                s_id_long = await ledger.credential_definition_id2schema_id(
                    f"55GkHamhTU1ZbTbV2ab9DE:3:CL:{s_id_short}:tag"
                )
                assert s_id_long == s_id_short

    @pytest.mark.asyncio
    async def test_rotate_did_keypair(self, ledger: IndyVdrLedger):
        async with ledger.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_public_did(SOV, ED25519)

        async with ledger:
            with mock.patch.object(
                ledger.pool_handle,
                "submit_request",
                mock.CoroutineMock(
                    side_effect=[
                        {"data": json.dumps({"seqNo": 1234})},
                        {"data": {"txn": {"data": {"role": "101", "alias": "Billy"}}}},
                        {"data": "ok"},
                    ]
                ),
            ):
                ledger.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
                await ledger.rotate_public_did_keypair()



from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import routes as test_module


class TestTrustpingRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_connections_send_ping(self):
        self.request.json = mock.CoroutineMock(return_value={"comment": "some comment"})
        self.request.match_info = {"conn_id": "dummy"}

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve,
            mock.patch.object(test_module, "Ping", mock.MagicMock()) as mock_ping,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as json_response,
        ):
            mock_ping.return_value = mock.MagicMock(_thread_id="dummy")
            mock_retrieve.return_value = mock.MagicMock(is_ready=True)
            result = await test_module.connections_send_ping(self.request)
            json_response.assert_called_once_with({"thread_id": "dummy"})
            assert result is json_response.return_value

    async def test_connections_send_ping_no_conn(self):
        self.request.json = mock.CoroutineMock(return_value={"comment": "some comment"})
        self.request.match_info = {"conn_id": "dummy"}

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_retrieve.side_effect = test_module.StorageNotFoundError()
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.connections_send_ping(self.request)

    async def test_connections_send_ping_not_ready(self):
        self.request.json = mock.CoroutineMock(return_value={"comment": "some comment"})
        self.request.match_info = {"conn_id": "dummy"}

        with (
            mock.patch.object(
                test_module.ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_retrieve.return_value = mock.MagicMock(is_ready=False)
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.connections_send_ping(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Configuration base classes."""

from abc import ABC, abstractmethod
from typing import Any, Iterator, Mapping, Optional, Type, TypeVar

from ..core.error import BaseError

InjectType = TypeVar("InjectType")


class ConfigError(BaseError):
    """A base exception for all configuration errors."""


class SettingsError(ConfigError):
    """The base exception raised by `BaseSettings` implementations."""


class BaseSettings(Mapping[str, Any]):
    """Base settings class."""

    @abstractmethod
    def get_value(self, *var_names, default: Optional[Any] = None) -> Any:
        """Fetch a setting.

        Args:
            var_names: A list of variable name alternatives
            default: The default value to return if none are defined

        Returns:
            The setting value, if defined, otherwise the default value

        """

    def get_bool(self, *var_names, default: Optional[bool] = None) -> Optional[bool]:
        """Fetch a setting as a boolean value.

        Args:
            var_names: A list of variable name alternatives
            default: The default value to return if none are defined
        """
        value = self.get_value(*var_names, default)
        if value is not None:
            value = bool(value and value not in ("false", "False", "0"))

        return value

    def get_int(self, *var_names, default: Optional[int] = None) -> Optional[int]:
        """Fetch a setting as an integer value.

        Args:
            var_names: A list of variable name alternatives
            default: The default value to return if none are defined
        """
        value = self.get_value(*var_names, default)
        if value is not None:
            value = int(value)

        return value

    def get_str(self, *var_names, default: Optional[str] = None) -> Optional[str]:
        """Fetch a setting as a string value.

        Args:
            var_names: A list of variable name alternatives
            default: The default value to return if none are defined
        """
        value = self.get_value(*var_names, default=default)
        if value is not None:
            value = str(value)

        return value

    @abstractmethod
    def __iter__(self) -> Iterator:
        """Iterate settings keys."""

    def __getitem__(self, index):
        """Fetch as an array index."""
        if not isinstance(index, str):
            raise TypeError(f"Index {index} must be a string")
        missing = object()
        result = self.get_value(index, default=missing)
        if result is missing:
            raise KeyError("Undefined index: {}".format(index))
        return result

    @abstractmethod
    def __len__(self):
        """Fetch the length of the mapping."""

    @abstractmethod
    def copy(self) -> "BaseSettings":
        """Produce a copy of the settings instance."""

    @abstractmethod
    def extend(self, other: Mapping[str, Any]) -> "BaseSettings":
        """Merge another mapping to produce a new settings instance."""

    @abstractmethod
    def to_dict(self) -> dict:
        """Return a dict of the settings instance."""

    def __repr__(self) -> str:
        """Provide a human readable representation of this object."""
        items = ("{}={}".format(k, self[k]) for k in self)
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))


class InjectionError(ConfigError):
    """The base exception raised by Injector and Provider implementations."""


class BaseInjector(ABC):
    """Base injector class."""

    @abstractmethod
    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, Any]] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            base_cls (Type[InjectType]): The base class to retrieve an instance of.
            settings (Optional[Mapping[str, Any]]): An optional mapping providing
                configuration to the provider.

        Returns:
            InjectType: An instance of the base class, or None.

        """

    @abstractmethod
    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, Any]] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """

    @abstractmethod
    def copy(self) -> "BaseInjector":
        """Produce a copy of the injector instance."""


class BaseProvider(ABC):
    """Base provider class."""

    def provide(self, settings: BaseSettings, injector: BaseInjector) -> Any:
        """Provide the object instance given a config and injector."""



"""Fixtures for vc holder tests."""

import pytest

from ..vc_record import VCRecord

VC_CONTEXT = "https://www.w3.org/2018/credentials/v1"
VC_TYPE = "https://www.w3.org/2018/credentials#VerifiableCredential"
VC_SUBJECT_ID = "did:example:ebfeb1f712ebc6f1c276e12ec21"
VC_PROOF_TYPE = "Ed25519Signature2018"
VC_ISSUER_ID = "https://example.edu/issuers/14"
VC_SCHEMA_ID = "https://example.org/examples/degree.json"
VC_GIVEN_ID = "http://example.edu/credentials/3732"


@pytest.fixture
def record():
    yield VCRecord(
        contexts=[
            VC_CONTEXT,
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        expanded_types=[
            VC_TYPE,
            "https://example.org/examples#UniversityDegreeCredential",
        ],
        schema_ids=[VC_SCHEMA_ID],
        issuer_id=VC_ISSUER_ID,
        subject_ids=[VC_SUBJECT_ID],
        proof_types=[VC_PROOF_TYPE],
        given_id=VC_GIVEN_ID,
        cred_tags={"tag": "value"},
        cred_value={
            "@context": [
                VC_CONTEXT,
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": VC_GIVEN_ID,
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": VC_ISSUER_ID,
            "identifier": "83627467",
            "name": "University Degree",
            "issuanceDate": "2010-01-01T19:53:24Z",
            "credentialSubject": {
                "id": VC_SUBJECT_ID,
                "givenName": "Cai",
                "familyName": "Leblanc",
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2021-05-07T08:50:17.626625",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..rubQvgig7cN-F6cYn_AJF1BCSaMpkoR517Ot_4pqwdJnQ-JwKXq6d6cNos5JR73E9WkwYISXapY0fYTIG9-fBA",
            },
        },
    )



"""Message and inner object type identifiers for present-proof protocol v1.0."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "4fae574c03f9f1013db30bf2c0c676b1122f7149/features/0037-present-proof"
)

# Message types
PRESENTATION_PROPOSAL = "present-proof/1.0/propose-presentation"
PRESENTATION_REQUEST = "present-proof/1.0/request-presentation"
PRESENTATION = "present-proof/1.0/presentation"
PRESENTATION_ACK = "present-proof/1.0/ack"
PRESENTATION_PROBLEM_REPORT = "present-proof/1.0/problem-report"

PROTOCOL_PACKAGE = "acapy_agent.protocols.present_proof.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        PRESENTATION_PROPOSAL: (
            f"{PROTOCOL_PACKAGE}.messages.presentation_proposal.PresentationProposal"
        ),
        PRESENTATION_REQUEST: (
            f"{PROTOCOL_PACKAGE}.messages.presentation_request.PresentationRequest"
        ),
        PRESENTATION: f"{PROTOCOL_PACKAGE}.messages.presentation.Presentation",
        PRESENTATION_ACK: (
            f"{PROTOCOL_PACKAGE}.messages.presentation_ack.PresentationAck"
        ),
        PRESENTATION_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.presentation_problem_report."
            "PresentationProblemReport"
        ),
    }
)

# Identifiers to use in attachment decorators
ATTACH_DECO_IDS = {
    PRESENTATION_REQUEST: "libindy-request-presentation-0",
    PRESENTATION: "libindy-presentation-0",
}

CONTROLLERS = DIDCommPrefix.qualify_all(
    {"present-proof/1.0": f"{PROTOCOL_PACKAGE}.controller.Controller"}
)



"""Abstract interfaces for VC holder implementations."""

from abc import ABC, abstractmethod
from typing import Mapping, Optional, Sequence

from .vc_record import VCRecord


class VCHolder(ABC):
    """Abstract base class for a verifiable credential holder."""

    @abstractmethod
    async def store_credential(self, cred: VCRecord):
        """Add a new VC record to the store.

        Args:
            cred: The VCRecord instance to store
        Raises:
            StorageDuplicateError: If the record_id is not unique

        """

    @abstractmethod
    async def retrieve_credential_by_id(self, record_id: str) -> VCRecord:
        """Fetch a VC record by its record ID.

        Raises:
            StorageNotFoundError: If the record is not found

        """

    @abstractmethod
    async def retrieve_credential_by_given_id(self, given_id: str) -> VCRecord:
        """Fetch a VC record by its given ID ('id' property).

        Raises:
            StorageNotFoundError: If the record is not found

        """

    @abstractmethod
    async def delete_credential(self, cred: VCRecord):
        """Remove a previously-stored VC record.

        Raises:
            StorageNotFoundError: If the record is not found

        """

    @abstractmethod
    def build_type_or_schema_query(self, uri_list: Sequence[str]) -> dict:
        """Build and return backend-specific type_or_schema_query.

        Args:
            uri_list: List of schema uri from input_descriptor
        """

    @abstractmethod
    def search_credentials(
        self,
        contexts: Sequence[str] = None,
        types: Sequence[str] = None,
        schema_ids: Sequence[str] = None,
        issuer_id: Optional[str] = None,
        subject_ids: Sequence[str] = None,
        proof_types: Sequence[str] = None,
        given_id: Optional[str] = None,
        tag_query: Optional[Mapping] = None,
    ) -> "VCRecordSearch":
        """Start a new VC record search.

        Args:
            contexts: An inclusive list of JSON-LD contexts to match
            types: An inclusive list of JSON-LD types to match
            schema_ids: An inclusive list of credential schema identifiers
            issuer_id: The ID of the credential issuer
            subject_ids: The IDs of any credential subjects all of which to match
            proof_types: The signature suite types used for the proof objects.
            given_id: The given id of the credential
            tag_query: A tag filter clause

        """

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        return "<{}>".format(self.__class__.__name__)


class VCRecordSearch(ABC):
    """A VC record search in progress."""

    @abstractmethod
    async def fetch(self, max_count: Optional[int] = None) -> Sequence[VCRecord]:
        """Fetch the next list of VC records from the store.

        Args:
            max_count: Max number of records to return. If not provided,
              defaults to the backend's preferred page size

        Returns:
            A list of `VCRecord` instances

        """

    async def close(self):
        """Dispose of the search query."""

    def __aiter__(self):
        """Async iterator magic method."""
        return IterVCRecordSearch(self)

    def __repr__(self) -> str:
        """Human readable representation of this instance."""
        return "<{}>".format(self.__class__.__name__)


class IterVCRecordSearch:
    """A generic record search async iterator."""

    def __init__(self, search: VCRecordSearch, page_size: Optional[int] = None):
        """Instantiate a new `IterVCRecordSearch` instance."""
        self._buffer = None
        self._page_size = page_size
        self._search = search

    async def __anext__(self):
        """Async iterator magic method."""
        if not self._buffer:
            self._buffer = await self._search.fetch(self._page_size) or []
        try:
            return self._buffer.pop(0)
        except IndexError:
            raise StopAsyncIteration
